@@ -2,6 +2,7 @@
 
 import hashlib
 import os
+from abc import ABC, abstractmethod
 from datetime import timedelta
 
 import flufl.lock
@@ -25,20 +26,74 @@ class LockError(DvcException):
     """Thrown when unable to acquire the lock for DVC repo."""
 
 
-class Lock:
+class LockBase(ABC):
+    @abstractmethod
+    def __init__(self, lockfile):
+        self._lockfile = lockfile
+
+    @property
+    def lockfile(self):
+        return self._lockfile
+
+    @abstractmethod
+    def lock(self):
+        pass
+
+    @abstractmethod
+    def unlock(self):
+        pass
+
+    @property
+    @abstractmethod
+    def is_locked(self):
+        pass
+
+    @abstractmethod
+    def __enter__(self):
+        pass
+
+    @abstractmethod
+    def __exit__(self, typ, value, tbck):
+        pass
+
+
+class LockNoop(LockBase):
+    def __init__(
+        self, *args, **kwargs
+    ):  # pylint: disable=super-init-not-called
+        self._lock = False
+
+    def lock(self):
+        self._lock = True
+
+    def unlock(self):
+        self._lock = False
+
+    @property
+    def is_locked(self):
+        return self._lock
+
+    def __enter__(self):
+        self.lock()
+
+    def __exit__(self, typ, value, tbck):
+        self.unlock()
+
+
+class Lock(LockBase):
     """Class for DVC repo lock.
 
     Uses zc.lockfile as backend.
     """
 
     def __init__(self, lockfile, friendly=False, **kwargs):
+        super().__init__(lockfile)
         self._friendly = friendly
-        self.lockfile = lockfile
         self._lock = None
 
     @property
     def files(self):
-        return [self.lockfile]
+        return [self._lockfile]
 
     def _do_lock(self):
         try:
@@ -51,7 +106,7 @@ class Lock:
                     )
                 ),
             ):
-                self._lock = zc.lockfile.LockFile(self.lockfile)
+                self._lock = zc.lockfile.LockFile(self._lockfile)
         except zc.lockfile.LockError:
             raise LockError(FAILED_TO_LOCK_MESSAGE)
 
@@ -76,7 +131,7 @@ class Lock:
         self.unlock()
 
 
-class HardlinkLock(flufl.lock.Lock):
+class HardlinkLock(flufl.lock.Lock, LockBase):
     """Class for DVC repo lock.
 
     Args:
@@ -90,6 +145,8 @@ class HardlinkLock(flufl.lock.Lock):
     ):  # pylint: disable=super-init-not-called
         import socket
 
+        super().__init__(lockfile)
+
         self._tmp_dir = tmp_dir
 
         # NOTE: this is basically Lock.__init__ copy-paste, except that
@@ -105,17 +162,12 @@ class HardlinkLock(flufl.lock.Lock):
         # [1] https://github.com/iterative/dvc/issues/2582
         self._hostname = socket.gethostname()
 
-        self._lockfile = lockfile
         self._lifetime = timedelta(days=365)  # Lock for good by default
         self._separator = flufl.lock.SEP
         self._set_claimfile()
         self._owned = True
         self._retry_errnos = []
 
-    @property
-    def lockfile(self):
-        return self._lockfile
-
     def lock(self):  # pylint: disable=arguments-differ
         try:
             super().lock(timedelta(seconds=DEFAULT_TIMEOUT))
