@@ -1,6 +1,7 @@
 from dvc.output.s3 import S3Output
-from dvc.remote.gs import GSRemote
+
+from ..tree.gs import GSTree
 
 
 class GSOutput(S3Output):
-    REMOTE = GSRemote
+    TREE_CLS = GSTree
