@@ -15,7 +15,7 @@ from dvc.exceptions import (
     OutputNotFoundError,
 )
 from dvc.path_info import PathInfo
-from dvc.repo.tree import RepoTree
+from dvc.tree.repo import RepoTree
 from dvc.utils.fs import path_isin
 
 from ..stage.exceptions import StageFileDoesNotExistError, StageNotFound
@@ -25,16 +25,30 @@ from .graph import check_acyclic, get_pipeline, get_pipelines
 logger = logging.getLogger(__name__)
 
 
+@contextmanager
+def lock_repo(repo):
+    # pylint: disable=protected-access
+    depth = getattr(repo, "_lock_depth", 0)
+    repo._lock_depth = depth + 1
+
+    try:
+        if depth > 0:
+            yield
+        else:
+            with repo.lock, repo.state:
+                repo._reset()
+                yield
+                # Graph cache is no longer valid after we release the repo.lock
+                repo._reset()
+    finally:
+        repo._lock_depth = depth
+
+
 def locked(f):
     @wraps(f)
     def wrapper(repo, *args, **kwargs):
-        with repo.lock, repo.state:
-            # pylint: disable=protected-access
-            repo._reset()
-            ret = f(repo, *args, **kwargs)
-            # Our graph cache is no longer valid after we release the repo.lock
-            repo._reset()
-            return ret
+        with lock_repo(repo):
+            return f(repo, *args, **kwargs)
 
     return wrapper
 
@@ -44,11 +58,11 @@ class Repo:
 
     from dvc.repo.add import add
     from dvc.repo.brancher import brancher
-    from dvc.repo.checkout import _checkout
+    from dvc.repo.checkout import checkout
     from dvc.repo.commit import commit
     from dvc.repo.destroy import destroy
     from dvc.repo.diff import diff
-    from dvc.repo.fetch import _fetch
+    from dvc.repo.fetch import fetch
     from dvc.repo.freeze import freeze, unfreeze
     from dvc.repo.gc import gc
     from dvc.repo.get import get
@@ -66,10 +80,17 @@ class Repo:
     from dvc.repo.status import status
     from dvc.repo.update import update
 
-    def __init__(self, root_dir=None, scm=None, rev=None):
+    def __init__(
+        self,
+        root_dir=None,
+        scm=None,
+        rev=None,
+        subrepos=False,
+        uninitialized=False,
+    ):
         from dvc.cache import Cache
         from dvc.data_cloud import DataCloud
-        from dvc.lock import make_lock
+        from dvc.lock import LockNoop, make_lock
         from dvc.repo.experiments import Experiments
         from dvc.repo.metrics import Metrics
         from dvc.repo.params import Params
@@ -80,65 +101,63 @@ class Repo:
         from dvc.tree.local import LocalTree
         from dvc.utils.fs import makedirs
 
-        if scm:
-            tree = scm.get_tree(rev)
+        try:
+            tree = scm.get_tree(rev) if rev else None
             self.root_dir = self.find_root(root_dir, tree)
-            self.scm = scm
-            self.tree = scm.get_tree(
-                rev, use_dvcignore=True, dvcignore_root=self.root_dir
-            )
-            self.state = StateNoop()
+            self.dvc_dir = os.path.join(self.root_dir, self.DVC_DIR)
+            self.tmp_dir = os.path.join(self.dvc_dir, "tmp")
+            makedirs(self.tmp_dir, exist_ok=True)
+        except NotDvcRepoError:
+            if not uninitialized:
+                raise
+            self.root_dir = SCM(root_dir or os.curdir).root_dir
+            self.dvc_dir = None
+            self.tmp_dir = None
+
+        tree_kwargs = dict(use_dvcignore=True, dvcignore_root=self.root_dir)
+        if scm:
+            self.tree = scm.get_tree(rev, **tree_kwargs)
         else:
-            root_dir = self.find_root(root_dir)
-            self.root_dir = os.path.abspath(os.path.realpath(root_dir))
-            self.tree = LocalTree(
-                self,
-                {"url": self.root_dir},
-                use_dvcignore=True,
-                dvcignore_root=self.root_dir,
-            )
+            self.tree = LocalTree(self, {"url": self.root_dir}, **tree_kwargs)
 
-        self.dvc_dir = os.path.join(self.root_dir, self.DVC_DIR)
         self.config = Config(self.dvc_dir, tree=self.tree)
+        no_scm = self.config["core"].get("no_scm", False)
+        self.scm = scm if scm else SCM(self.root_dir, no_scm=no_scm)
 
-        if not scm:
-            no_scm = self.config["core"].get("no_scm", False)
-            self.scm = SCM(self.root_dir, no_scm=no_scm)
-
-        self.tmp_dir = os.path.join(self.dvc_dir, "tmp")
-        self.index_dir = os.path.join(self.tmp_dir, "index")
-        makedirs(self.index_dir, exist_ok=True)
-
-        hardlink_lock = self.config["core"].get("hardlink_lock", False)
-        self.lock = make_lock(
-            os.path.join(self.tmp_dir, "lock"),
-            tmp_dir=self.tmp_dir,
-            hardlink_lock=hardlink_lock,
-            friendly=True,
-        )
+        # used by RepoTree to determine if it should traverse subrepos
+        self.subrepos = subrepos
 
         self.cache = Cache(self)
         self.cloud = DataCloud(self)
 
-        if not scm:
+        if scm or not self.dvc_dir:
+            self.lock = LockNoop()
+            self.state = StateNoop()
+        else:
+            self.lock = make_lock(
+                os.path.join(self.tmp_dir, "lock"),
+                tmp_dir=self.tmp_dir,
+                hardlink_lock=self.config["core"].get("hardlink_lock", False),
+                friendly=True,
+            )
+
             # NOTE: storing state and link_state in the repository itself to
             # avoid any possible state corruption in 'shared cache dir'
             # scenario.
-            self.state = State(self.cache.local)
+            self.state = State(self)
+            self.stage_cache = StageCache(self)
 
-        self.stage_cache = StageCache(self)
+            try:
+                self.experiments = Experiments(self)
+            except NotImplementedError:
+                self.experiments = None
+
+            self._ignore()
 
         self.metrics = Metrics(self)
         self.plots = Plots(self)
         self.params = Params(self)
 
-        try:
-            self.experiments = Experiments(self)
-        except NotImplementedError:
-            self.experiments = None
-
-        self._ignore()
-
     @property
     def tree(self):
         return self._tree
@@ -526,12 +545,6 @@ class Repo:
         """
         return self._collect_stages()
 
-    @cached_property
-    def plot_templates(self):
-        from .plots.template import PlotTemplates
-
-        return PlotTemplates(self.dvc_dir)
-
     def _collect_stages(self):
         stages = []
         outs = set()
@@ -581,19 +594,20 @@ class Repo:
         path_parts = os.path.normpath(path).split(os.path.sep)
         return self.DVC_DIR in path_parts
 
+    @cached_property
+    def repo_tree(self):
+        return RepoTree(self, subrepos=self.subrepos, fetch=True)
+
     @contextmanager
     def open_by_relpath(self, path, remote=None, mode="r", encoding=None):
         """Opens a specified resource as a file descriptor"""
 
-        tree = RepoTree(self, stream=True)
-        path = os.path.join(self.root_dir, path)
+        tree = RepoTree(self, stream=True, subrepos=True)
+        path = PathInfo(self.root_dir) / path
         try:
             with self.state:
                 with tree.open(
-                    os.path.join(self.root_dir, path),
-                    mode=mode,
-                    encoding=encoding,
-                    remote=remote,
+                    path, mode=mode, encoding=encoding, remote=remote,
                 ) as fobj:
                     yield fobj
         except FileNotFoundError as exc:
@@ -604,14 +618,6 @@ class Repo:
     def close(self):
         self.scm.close()
 
-    @locked
-    def checkout(self, *args, **kwargs):
-        return self._checkout(*args, **kwargs)
-
-    @locked
-    def fetch(self, *args, **kwargs):
-        return self._fetch(*args, **kwargs)
-
     def _reset(self):
         self.__dict__.pop("graph", None)
         self.__dict__.pop("stages", None)
